"""Make VDSGenerator easy to import."""
from .vdsgenerator import VDSGenerator
from .subframevdsgenerator import SubFrameVDSGenerator
from .framevdsgenerator import FrameVDSGenerator

__all__ = ["FrameVDSGenerator", "SubFrameVDSGenerator"]



from h5py._hl.group import Group
from h5py._hl.base import phil
from h5py import h5p, h5s, h5d, h5t

import numpy

from copy import deepcopy as copy


class VGroup(Group):

    def create_virtual_dataset(self, VMlist, fillvalue=None):
        """
        Creates the virtual dataset from a list of virtual maps, any gaps are filled with a specified fill value.

        VMlist
            (List) A list of the the VirtualMaps between the source and target datasets. At least one is required.
        fillvalue
            (Scalar) Use this value for uninitialized parts of the dataset.
        """

        if not VMlist:
            raise ValueError(
                "create_virtual_dataset requires at least one virtual map to construct output.")

        if not isinstance(VMlist, (tuple, list)):
            VMlist = [VMlist]
        with phil:
            dcpl = h5p.create(h5p.DATASET_CREATE)
            dcpl.set_fill_value(numpy.array([fillvalue]))
            sh = VMlist[0].target.shape
            virt_dspace = h5s.create_simple(sh, VMlist[
                0].target.maxshape)  # create the virtual dataspace
            for VM in VMlist:
                virt_start_idx = tuple([ix.start for ix in VM.target.slice_list])
                virt_stride_index = tuple([ix.step for ix in VM.target.slice_list])
                if any(ix == h5s.UNLIMITED for ix in VM.target.maxshape):
                    count_idx = [1, ] * len(virt_stride_index)
                    unlimited_index = VM.target.maxshape.index(h5s.UNLIMITED)
                    count_idx[unlimited_index] = h5s.UNLIMITED
                    count_idx = tuple(count_idx)
                else:
                    count_idx = (1,) * len(virt_stride_index)
                virt_dspace.select_hyperslab(start=virt_start_idx,
                                             count=count_idx,
                                             stride=virt_stride_index,
                                             block=VM.block_shape)
                dcpl.set_virtual(virt_dspace, str.encode(str(VM.src.path)), str.encode(str(VM.src.key)),
                                 VM.src_dspace)
            dset = h5d.create(self.id, name=str.encode(str(VM.target.key)),
                              tid=h5t.py_create(VM.dtype, logical=1),
                              space=virt_dspace, dcpl=dcpl)

        return dset


class DatasetContainer(object):
    def __init__(self, path, key, shape, dtype=None, maxshape=None):
        """
        This is an object that looks like a dataset, but it not. It allows the user to specify the maps based on lazy indexing,
        which is natural, but without needing to load the data.

        path
            This is the full path to the file on disk.

        key
            This is the key to the entry inside the hdf5 file.

        shape
            The shape of the data. We specify this by hand because it is a lot faster than getting it from the source file.

        dtype
            The data type. For the source we specify this because it is faster than getting from the file. For the target,
            we can specify it to be different to the source.
        """
        self.path = path
        self.key = key
        self.shape = shape
        self.slice_list = [slice(0, ix, 1) for ix in
                           self.shape]  # if we don't slice, we want the whole array
        if maxshape is None:
            self.maxshape = shape
        else:
            self.maxshape = tuple(
                [h5s.UNLIMITED if ix is None else ix for ix in maxshape])

    def _parse_slicing(self, key):
        """
        parses the __get_item__ key to get useful slicing information
        """
        tmp = copy(self)
        rank = len(self.shape)
        if (rank - len(key)) < 0:
            raise IndexError('Index rank is greater than dataset rank')
        if isinstance(key[0], tuple):  # sometimes this is needed. odd
            key = key[0]
        key = list(key)
        key = [slice(ix, ix + 1, 1) if isinstance(ix, (int, float)) else ix for
               ix in key]

        # now let's parse ellipsis
        ellipsis_test = [ix == Ellipsis for ix in key]
        if sum(ellipsis_test) > 1:
            raise ValueError("Only use of one Ellipsis(...) supported.")
        if not any(ellipsis_test):
            tmp.slice_list[:len(key)] = key
        elif any(ellipsis_test) and (len(key) is not 1):
            ellipsis_idx = ellipsis_test.index(True)
            ellipsis_idx_back = ellipsis_test[::-1].index(True)
            tmp.slice_list[0:ellipsis_idx] = key[0:ellipsis_idx]
            if ellipsis_idx_back >= ellipsis_idx:  # edge case
                tmp.slice_list[(-ellipsis_idx_back):] = key[
                                                        (-ellipsis_idx_back):]

        new_shape = []
        for ix, sl in enumerate(tmp.slice_list):
            step = 1 if sl.step is None else sl.step
            if step > 0:
                start = 0 if sl.start is None else sl.start  # parse for Nones
                stop = self.shape[ix] if sl.stop is None else sl.stop
                start = self.shape[ix] + start if start < 0 else start
                stop = self.shape[ix] + stop if stop < 0 else stop
                if start < stop:
                    new_shape.append((stop - start + step - 1) / step)
                else:
                    new_shape.append(0)

            elif step < 0:
                stop = 0 if sl.stop is None else sl.stop  # parse for Nones
                start = self.shape[ix] if sl.start is None else sl.start

                start = self.shape[ix] + start if start < 0 else start
                stop = self.shape[ix] + stop if stop < 0 else stop

                if start > stop:  # this gets the same behaviour as numpy array
                    new_shape.append((start - stop - step - 1) / -step)
                else:
                    new_shape.append(0)
            elif step == 0:
                raise IndexError("A step of 0 is not valid")
            tmp.slice_list[ix] = slice(start, stop, step)
        tmp.shape = tuple(new_shape)
        return tmp


class VirtualSource(DatasetContainer):
    """
    A container for the source information. This is similar to a virtual target, but the shape information changes with slicing.
    This does not happen with VirtualTarget since it is the source that ultimately set's the block shape.
    """
    def __getitem__(self, *key):
        tmp = self._parse_slicing(key)
        return tmp


class VirtualTarget(DatasetContainer):
    """
    A container for the target information. This is similar to a virtual source, but the shape information does not change with slicing.
    This does not happen with VirtualSource since it is the source that ultimately set's the block shape so it must change on slicing.
    """
    def __getitem__(self, *key):
        tmp = self._parse_slicing(key)
        tmp.shape = self.shape
        return tmp


class VirtualMap(object):
    def __init__(self, virtual_source, virtual_target, dtype):
        """
        The idea of this class is to specify the mapping between the source and target files.
        Since data type casting is supported by VDS, we include this here.
            virtual_source
                A DatasetContainer object containing all the useful information about the source file for this map.

            virtual_target
                A DatasetContainer object containing all the useful information about the source file for this map.
            dtype
                The type of the final output dataset.

        """
        self.src = virtual_source[...]
        self.dtype = dtype
        self.target = virtual_target[...]
        self.block_shape = None
        # if the rank of the two datasets is not the same, pad with singletons. This isn't necessarily the best way to do this!
        rank_def = len(self.target.shape) - len(self.src.shape)
        if rank_def > 0:
            if len(self.src.shape) == 1:
                pass
            else:
                self.block_shape = (1,) * rank_def + self.src.shape
        elif rank_def < 0:
            # This might be pathological.
            if len(self.target.shape) == 1:
                pass
            else:
                self.block_shape = (1,) * rank_def + self.target.shape
        else:
            self.block_shape = self.src.shape

        self.src_dspace = h5s.create_simple(self.src.shape, self.src.maxshape)

        start_idx = tuple([ix.start for ix in self.src.slice_list])
        stride_idx = tuple([ix.step for ix in self.src.slice_list])

        if any(ix == h5s.UNLIMITED for ix in self.src.maxshape):
            count_idx = [1, ] * len(stride_idx)
            unlimited_index = self.src.maxshape.index(h5s.UNLIMITED)
            count_idx[unlimited_index] = h5s.UNLIMITED
            count_idx = tuple(count_idx)
            bs = list(self.block_shape)
            bs[unlimited_index] = 1
            self.block_shape = tuple(bs)
        else:
            count_idx = (1,) * len(stride_idx)
        self.src_dspace.select_hyperslab(start=start_idx, count=count_idx, stride=stride_idx, block=self.block_shape)



import unittest
from pkg_resources import require

require("mock")
from mock import MagicMock, patch, call

from vdsgen import app

app_patch_path = "vdsgen.app"
parser_patch_path = app_patch_path + ".ArgumentParser"
VDSGenerator_patch_path = app_patch_path + ".VDSGenerator"
SubFrameVDSGenerator_patch_path = app_patch_path + ".SubFrameVDSGenerator"


class ParseArgsTest(unittest.TestCase):

    @patch(SubFrameVDSGenerator_patch_path)
    @patch(VDSGenerator_patch_path)
    @patch(app_patch_path + '.ArgumentDefaultsHelpFormatter')
    @patch(parser_patch_path)
    def test_parser(self, parser_init_mock, formatter_mock, gen_mock,
                    subframegen_mock):
        parser_mock = parser_init_mock.return_value
        add_mock = parser_mock.add_argument
        add_group_mock = parser_mock.add_argument_group
        add_exclusive_group_mock = parser_mock.add_mutually_exclusive_group
        parse_mock = parser_mock.parse_args
        parse_mock.return_value = MagicMock(empty=False, files=None)
        empty_mock = MagicMock()
        other_mock = MagicMock()
        add_group_mock.side_effect = [empty_mock, other_mock]
        exclusive_group_mock = add_exclusive_group_mock.return_value
        expected_message = """
-------------------------------------------------------------------------------
A script to create a virtual dataset composed of multiple raw HDF5 files.

The minimum required arguments are <path> and either -p <prefix> or -f <files>.

For example:

 > ../vdsgen/app.py /scratch/images -p stripe_
 > ../vdsgen/app.py /scratch/images -f stripe_1.hdf5 stripe_2.hdf5

You can create an empty VDS, for raw files that don't exist yet, with the -e
flag; you will then need to provide --shape and --data_type, though defaults
are provided for these.
-------------------------------------------------------------------------------
"""

        args = app.parse_args()

        parser_init_mock.assert_called_once_with(
            usage=expected_message,
            formatter_class=formatter_mock)
        add_exclusive_group_mock.assert_called_with(required=True)
        exclusive_group_mock.add_argument.assert_has_calls(
            [call("-p", "--prefix", type=str, default=None, dest="prefix",
                  help="Prefix of files to search <path> for - e.g 'stripe_' "
                       "to combine 'stripe_1.hdf5' and 'stripe_2.hdf5'."),
             call("-f", "--files", nargs="*", type=str, default=None,
                  dest="files",
                  help="Explicit names of raw files in <path>.")])

        add_mock.assert_called_with(
            "path", type=str, help="Root folder of source files and VDS.")

        add_group_mock.assert_has_calls([call()] * 2)
        empty_mock.add_argument.assert_has_calls(
            [call("-e", "--empty", action="store_true", dest="empty",
                  help="Make empty VDS pointing to datasets "
                       "that don't exist yet."),
             call("--shape", type=int, nargs="*", default=[1, 256, 2048],
                  dest="shape",
                  help="Shape of dataset - 'frames height width', where "
                       "frames is N dimensional."),
             call("-t", "--data_type", type=str, default="uint16",
                  dest="data_type", help="Data type of raw datasets.")])
        other_mock.add_argument.assert_has_calls(
            [call("-o", "--output", type=str, default=None, dest="output",
                  help="Output file name. If None then generated as input "
                       "file prefix with vds suffix."),
             call("-s", "--stripe_spacing", type=int, dest="stripe_spacing",
                  default=subframegen_mock.stripe_spacing,
                  help="Spacing between two stripes in a module."),
             call("-m", "--module_spacing", type=int, dest="module_spacing",
                  default=subframegen_mock.module_spacing,
                  help="Spacing between two modules."),
             call("-F", "--fill_value", type=int, dest="fill_value", default=0,
                  help="Fill value for spacing."),
             call("--source_node", type=str, dest="source_node",
                  default=gen_mock.source_node,
                  help="Data node in source HDF5 files."),
             call("--target_node", type=str,
                  default=gen_mock.target_node, dest="target_node",
                  help="Data node in VDS file."),
             call("--mode", default="sub-frames", dest="mode",
                  help="Type of raw datasets  sub-frames: ND datasets containing sub_frames of full image  frames:     1D datasets containing interspersed frames",
                  type=str),
             call("-l", "--log_level", type=int, dest="log_level",
                  default=gen_mock.log_level,
                  help="Logging level (off=3, info=2, debug=1).")])

        parse_mock.assert_called_once_with()
        self.assertEqual(parse_mock.return_value, args)

    @patch(parser_patch_path + '.error')
    @patch(parser_patch_path + '.parse_args',
           return_value=MagicMock(empty=True, files=None))
    def test_empty_and_not_files_then_error(self, parse_mock, error_mock):

        app.parse_args()

        error_mock.assert_called_once_with(
            "To make an empty VDS you must explicitly define --files for the "
            "eventual raw datasets.")

    @patch(parser_patch_path + '.error')
    @patch(parser_patch_path + '.parse_args',
           return_value=MagicMock(empty=True, files=["file"]))
    def test_only_one_file_then_error(self, parse_mock, error_mock):

        app.parse_args()

        error_mock.assert_called_once_with(
            "Must define at least two files to combine.")


class MainTest(unittest.TestCase):

    @patch(SubFrameVDSGenerator_patch_path)
    @patch(app_patch_path + '.parse_args',
           return_value=MagicMock(
               path="/test/path", empty=True,
               prefix=None, files=["file1.hdf5", "file2.hdf5"], output="vds",
               shape=(3, 256, 2048), data_type="int16",
               source_node="data", target_node="full_frame",
               stripe_spacing=3, module_spacing=127, fill_value=-1,
               mode="sub-frames",
               log_level=2))
    def test_main_empty(self, parse_mock, init_mock):
        gen_mock = init_mock.return_value
        args_mock = parse_mock.return_value

        app.main()

        parse_mock.assert_called_once_with()
        init_mock.assert_called_once_with(
            args_mock.path,
            prefix=args_mock.prefix, files=args_mock.files,
            output=args_mock.output,
            source=dict(shape=args_mock.shape, dtype=args_mock.data_type),
            source_node=args_mock.source_node,
            target_node=args_mock.target_node,
            stripe_spacing=args_mock.stripe_spacing,
            module_spacing=args_mock.module_spacing,
            fill_value=-1,
            log_level=args_mock.log_level)

        gen_mock.generate_vds.assert_called_once_with()

    @patch("os.path.isfile", return_value=True)
    @patch(SubFrameVDSGenerator_patch_path)
    @patch(app_patch_path + '.parse_args',
           return_value=MagicMock(
               path="/test/path", empty=False,
               prefix=None, files=["file1.hdf5", "file2.hdf5"], output="vds",
               frames=3, height=256, width=2048, data_type="int16",
               source_node="data", target_node="full_frame",
               stripe_spacing=3, module_spacing=127, fill_value=-1,
               mode="sub-frames",
               log_level=2))
    def test_main_not_empty(self, parse_mock, generate_mock, _):
        args_mock = parse_mock.return_value

        app.main()

        parse_mock.assert_called_once_with()
        generate_mock.assert_called_once_with(
            args_mock.path,
            prefix=args_mock.prefix, output="vds", files=args_mock.files,
            source=None,
            source_node=args_mock.source_node,
            stripe_spacing=args_mock.stripe_spacing,
            target_node=args_mock.target_node,
            module_spacing=args_mock.module_spacing,
            fill_value=-1,
            log_level=args_mock.log_level)



#!/bin/env dls-python
"""A class for generating virtual dataset frames from sub-frames."""

import h5py as h5

from .vdsgenerator import VDSGenerator, SourceMeta
from .group import VirtualSource, VirtualTarget, VirtualMap


class FrameVDSGenerator(VDSGenerator):

    """A class to generate Virtual Dataset frames from sub-frames."""

    def __init__(self, path, prefix=None, files=None, output=None, source=None,
                 source_node=None, target_node=None, fill_value=None,
                 log_level=None):
        """
        Args:
            path(str): Root folder to find raw files and create VDS
            prefix(str): Prefix of HDF5 files to generate from
                e.g. image_ for image_1.hdf5, image_2.hdf5, image_3.hdf5
            files(list(str)): List of HDF5 files to generate from
            output(str): Name of VDS file.
            source(dict): Height, width, data_type and frames for source data
                Provide this to create a VDS for raw files that don't exist yet
            source_node(str): Data node in source HDF5 files
            target_node(str): Data node in VDS file
            fill_value(int): Fill value for spacing
            log_level(int): Logging level (off=3, info=2, debug=1) -
                Default is info

        """
        self.total_frames = 0

        super(FrameVDSGenerator, self).__init__(
            path, prefix, files, output, source, source_node, target_node,
            fill_value, log_level)

    def process_source_datasets(self):
        """Grab data from the given HDF5 files and check for consistency.

        Returns:
            Source: Number of datasets and the attributes of them (frames,
                height width and data type)

        """
        data = self.grab_metadata(self.datasets[0])
        self.total_frames = data["frames"][0]
        for dataset in self.datasets[1:]:
            temp_data = self.grab_metadata(dataset)
            self.total_frames += temp_data["frames"][0]
            for attribute, value in data.items():
                if attribute != "frames" and temp_data[attribute] != value:
                    raise ValueError("Files have mismatched "
                                     "{}".format(attribute))

        source = SourceMeta(frames=data['frames'],
                            height=data['height'], width=data['width'],
                            dtype=data['dtype'])

        self.logger.debug("Source metadata retrieved:\n"
                          "  Frames: %s\n"
                          "  Height: %s\n"
                          "  Width: %s\n"
                          "  Data Type: %s", self.total_frames, *source[1:])
        return source

    def create_vds_maps(self, source_meta):
        """Create a list of VirtualMaps of raw data to the VDS.

        Args:
            source_meta(SourceMeta): Source attributes

        Returns:
            list(VirtualMap): Maps describing links between raw data and VDS

        """
        source_dims = (source_meta.height, source_meta.width)
        target_shape = (self.total_frames,
                        source_meta.height, source_meta.width)
        self.logger.debug("VDS metadata:\n"
                          "  Shape: %s\n", target_shape)

        vds = VirtualTarget(self.output_file, self.target_node,
                            shape=target_shape)

        total_datasets = len(self.datasets)

        map_list = []
        for dataset_idx, dataset in enumerate(self.datasets):
            dataset_frames = h5.File(dataset)[self.source_node].shape[0]
            source = VirtualSource(dataset, self.source_node,
                                   shape=(dataset_frames,)+source_dims)

            for frame_idx in range(dataset_frames):

                # Hyperslab: Frame[frame_idx],
                #            Full slice for height and width
                source_hyperslab = tuple([slice(frame_idx, frame_idx + 1)] +
                                         [self.FULL_SLICE, self.FULL_SLICE])
                v_source = source[source_hyperslab]

                # Hyperslab: Single frame based on indexes,
                #            Full slice for height and width
                target_idx = dataset_idx + total_datasets * frame_idx
                vds_hyperslab = tuple([slice(target_idx, target_idx + 1)] +
                                      [self.FULL_SLICE, self.FULL_SLICE])
                v_target = vds[vds_hyperslab]

                v_map = VirtualMap(v_source, v_target,
                                   dtype=source_meta.dtype)

                self.logger.debug("Mapping frame %s of %s to %s of %s.",
                                  frame_idx, dataset.split("/")[-1],
                                  target_idx, self.name)
                map_list.append(v_map)

        return map_list



import os
import sys
import unittest

from pkg_resources import require

require("mock")
from mock import MagicMock, patch, call

from vdsgen import vdsgenerator
from vdsgen.subframevdsgenerator import SubFrameVDSGenerator

vdsgen_patch_path = "vdsgen.subframevdsgenerator"
VDSGenerator_patch_path = vdsgen_patch_path + ".VDSGenerator"
h5py_patch_path = "h5py"

sys.path.append(os.path.join(os.path.dirname(__file__), "..", "..", "h5py"))


class SubFrameVDSGeneratorTester(SubFrameVDSGenerator):
    """A version of SubFrameVDSGenerator without initialisation.

    For testing single methods of the class. Must have required attributes
    passed before calling testee function.

    """

    def __init__(self, **kwargs):
        for attribute, value in kwargs.items():
            self.__setattr__(attribute, value)
        self.logger = MagicMock()


class FrameVDSGeneratorInitTest(unittest.TestCase):

    @patch(VDSGenerator_patch_path + '.__init__')
    def test_super_called(self, super_mock):
        SubFrameVDSGenerator("/test/path", prefix="stripe_")

        super_mock.assert_called_once_with("/test/path", "stripe_",
                                           *[None]*7)


class SimpleFunctionsTest(unittest.TestCase):

    @patch(VDSGenerator_patch_path + '.grab_metadata',
           return_value=dict(frames=(3,), height=256, width=2048,
                             dtype="uint16"))
    def test_process_source_datasets_given_valid_data(self, grab_mock):
        gen = SubFrameVDSGeneratorTester(
            datasets=["stripe_1.h5", "stripe_2.h5"])
        expected_source = vdsgenerator.SourceMeta(
            frames=(3,), height=256, width=2048, dtype="uint16")

        source = gen.process_source_datasets()

        grab_mock.assert_has_calls([call("stripe_1.h5"), call("stripe_2.h5")])
        self.assertEqual(expected_source, source)

    @patch(VDSGenerator_patch_path + '.grab_metadata',
           side_effect=[dict(frames=3, height=256, width=2048, dtype="uint16"),
                        dict(frames=4, height=256, width=2048,
                             dtype="uint16")])
    def test_process_source_datasets_given_mismatched_data(self, grab_mock):
        gen = SubFrameVDSGeneratorTester(
            datasets=["stripe_1.h5", "stripe_2.h5"])

        with self.assertRaises(ValueError):
            gen.process_source_datasets()

        grab_mock.assert_has_calls([call("stripe_1.h5"), call("stripe_2.h5")])

    def test_construct_vds_spacing(self):
        gen = SubFrameVDSGeneratorTester(
            datasets=[""] * 6, stripe_spacing=10, module_spacing=100)
        expected_spacing = [10, 100, 10, 100, 10, 0]

        spacing = gen.construct_vds_spacing()

        self.assertEqual(expected_spacing, spacing)

    @patch(vdsgen_patch_path + '.VirtualMap')
    @patch(vdsgen_patch_path + '.VirtualSource')
    @patch(vdsgen_patch_path + '.VirtualTarget')
    def test_create_vds_maps(self, target_mock, source_mock, map_mock):
        gen = SubFrameVDSGeneratorTester(
            output_file="/test/path/vds.hdf5",
            stripe_spacing=10, module_spacing=100,
            target_node="full_frame", source_node="data",
            datasets=["source"] * 6, name="vds.hdf5")
        source = vdsgenerator.SourceMeta(
            frames=(3,), height=256, width=2048, dtype="uint16")

        map_list = gen.create_vds_maps(source)

        target_mock.assert_called_once_with("/test/path/vds.hdf5",
                                            "full_frame",
                                            shape=(3, 1766, 2048))
        source_mock.assert_has_calls([call("source", "data",
                                           shape=(3, 256, 2048))] * 6)
        # TODO: Improve this assert by passing numpy arrays to check slicing
        map_mock.assert_has_calls([
            call(source_mock.return_value,
                 target_mock.return_value.__getitem__.return_value,
                 dtype="uint16")] * 6)
        self.assertEqual([map_mock.return_value] * 6, map_list)



import sys
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import logging

from .vdsgenerator import VDSGenerator
from .framevdsgenerator import FrameVDSGenerator
from .subframevdsgenerator import SubFrameVDSGenerator

help_message = """
-------------------------------------------------------------------------------
A script to create a virtual dataset composed of multiple raw HDF5 files.

The minimum required arguments are <path> and either -p <prefix> or -f <files>.

For example:

 > ../vdsgen/app.py /scratch/images -p stripe_
 > ../vdsgen/app.py /scratch/images -f stripe_1.hdf5 stripe_2.hdf5

You can create an empty VDS, for raw files that don't exist yet, with the -e
flag; you will then need to provide --shape and --data_type, though defaults
are provided for these.
-------------------------------------------------------------------------------
"""


def parse_args():
    """Parse command line arguments."""
    parser = ArgumentParser(usage=help_message,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "path", type=str, help="Root folder of source files and VDS.")

    # Definition of file names in <path> - Common prefix or explicit list
    file_definition = parser.add_mutually_exclusive_group(required=True)
    file_definition.add_argument(
        "-p", "--prefix", type=str, default=None, dest="prefix",
        help="Prefix of files to search <path> for - e.g 'stripe_' to combine "
             "'stripe_1.hdf5' and 'stripe_2.hdf5'.")
    file_definition.add_argument(
        "-f", "--files", nargs="*", type=str, default=None, dest="files",
        help="Explicit names of raw files in <path>.")

    # Arguments required to allow VDS to be created before raw files exist
    empty_vds = parser.add_argument_group()
    empty_vds.add_argument(
        "-e", "--empty", action="store_true", dest="empty",
        help="Make empty VDS pointing to datasets that don't exist yet.")
    empty_vds.add_argument(
        "--shape", type=int, nargs="*", default=[1, 256, 2048], dest="shape",
        help="Shape of dataset - 'frames height width', where frames is N "
             "dimensional.")
    empty_vds.add_argument(
        "-t", "--data_type", type=str, default="uint16", dest="data_type",
        help="Data type of raw datasets.")

    # Arguments to override defaults - each is atomic
    other_args = parser.add_argument_group()
    other_args.add_argument(
        "-o", "--output", type=str, default=None, dest="output",
        help="Output file name. If None then generated as input file prefix "
             "with vds suffix.")
    other_args.add_argument(
        "-s", "--stripe_spacing", type=int, dest="stripe_spacing",
        default=SubFrameVDSGenerator.stripe_spacing,
        help="Spacing between two stripes in a module.")
    other_args.add_argument(
        "-m", "--module_spacing", type=int, dest="module_spacing",
        default=SubFrameVDSGenerator.module_spacing,
        help="Spacing between two modules.")
    other_args.add_argument(
        "-F", "--fill_value", type=int, dest="fill_value", default=0,
        help="Fill value for spacing.")
    other_args.add_argument(
        "--source_node", type=str, dest="source_node",
        default=VDSGenerator.source_node,
        help="Data node in source HDF5 files.")
    other_args.add_argument(
        "--target_node", type=str, dest="target_node",
        default=VDSGenerator.target_node, help="Data node in VDS file.")
    other_args.add_argument(
        "--mode", type=str, dest="mode", default="sub-frames",
        help="Type of raw datasets"
             "  sub-frames: ND datasets containing sub_frames of full image"
             "  frames:     1D datasets containing interspersed frames")
    other_args.add_argument(
        "-l", "--log_level", type=int, dest="log_level",
        default=VDSGenerator.log_level,
        help="Logging level (off=3, info=2, debug=1).")

    args = parser.parse_args()
    args.shape = tuple(args.shape)

    if args.empty and args.files is None:
        parser.error(
            "To make an empty VDS you must explicitly define --files for the "
            "eventual raw datasets.")
    if args.files is not None and len(args.files) < 2:
        parser.error("Must define at least two files to combine.")

    return args


def main():
    """Run program."""
    logging.basicConfig(
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")

    args = parse_args()

    if args.empty:
        source_metadata = dict(shape=args.shape, dtype=args.data_type)
    else:
        source_metadata = None

    if args.mode == "frames":
        gen = FrameVDSGenerator(args.path,
                                prefix=args.prefix, files=args.files,
                                output=args.output,
                                source=source_metadata,
                                source_node=args.source_node,
                                target_node=args.target_node,
                                fill_value=args.fill_value,
                                log_level=args.log_level)
    elif args.mode == "sub-frames":
        gen = SubFrameVDSGenerator(args.path,
                                   prefix=args.prefix, files=args.files,
                                   output=args.output,
                                   source=source_metadata,
                                   source_node=args.source_node,
                                   target_node=args.target_node,
                                   stripe_spacing=args.stripe_spacing,
                                   module_spacing=args.module_spacing,
                                   fill_value=args.fill_value,
                                   log_level=args.log_level)
    else:
        raise NotImplementedError("Invalid VDS mode. Must be either frames or "
                                  "sub-frames.")

    gen.generate_vds()


if __name__ == "__main__":
    sys.exit(main())



#!/bin/env dls-python
"""A class for generating virtual dataset frames from sub-frames."""

from .vdsgenerator import VDSGenerator, SourceMeta
from .group import VirtualSource, VirtualTarget, VirtualMap


class SubFrameVDSGenerator(VDSGenerator):

    """A class to generate Virtual Dataset frames from sub-frames."""

    # Default Values
    stripe_spacing = 10  # Pixel spacing between stripes in a module
    module_spacing = 10  # Pixel spacing between modules

    def __init__(self, path, prefix=None, files=None, output=None, source=None,
                 source_node=None, target_node=None, fill_value=None,
                 stripe_spacing=None, module_spacing=None,
                 log_level=None):
        """
        Args:
            path(str): Root folder to find raw files and create VDS
            prefix(str): Prefix of HDF5 files to generate from
                e.g. image_ for image_1.hdf5, image_2.hdf5, image_3.hdf5
            files(list(str)): List of HDF5 files to generate from
            output(str): Name of VDS file.
            source(dict): Height, width, data_type and frames for source data
                Provide this to create a VDS for raw files that don't exist yet
            source_node(str): Data node in source HDF5 files
            target_node(str): Data node in VDS file
            fill_value(int): Fill value for spacing
            stripe_spacing(int): Spacing between stripes in module
            module_spacing(int): Spacing between modules
            log_level(int): Logging level (off=3, info=2, debug=1) -
                Default is info

        """
        super(SubFrameVDSGenerator, self).__init__(
            path, prefix, files, output, source, source_node, target_node,
            fill_value, log_level)

        # Overwrite default values with arguments, if given
        if stripe_spacing is not None:
            self.stripe_spacing = stripe_spacing
        if module_spacing is not None:
            self.module_spacing = module_spacing

    def process_source_datasets(self):
        """Grab data from the given HDF5 files and check for consistency.

        Returns:
            Source: Number of datasets and the attributes of them (frames,
                height width and data type)

        """
        data = self.grab_metadata(self.datasets[0])
        frames = [data["frames"]]
        for dataset in self.datasets[1:]:
            temp_data = self.grab_metadata(dataset)
            frames.append(temp_data["frames"])
            for attribute, value in data.items():
                if temp_data[attribute] != value:
                    raise ValueError("Files have mismatched "
                                     "{}".format(attribute))

        source = SourceMeta(frames=data['frames'], height=data['height'],
                            width=data['width'], dtype=data['dtype'])

        self.logger.debug("Source metadata retrieved:\n"
                          "  Frames: %s\n"
                          "  Height: %s\n"
                          "  Width: %s\n"
                          "  Data Type: %s", frames, *source[1:])
        return source

    def construct_vds_spacing(self):
        """Construct list of spacing between each stripe.

        Returns:
            VDS: Dataset spacing of virtual data set

        """
        stripes = len(self.datasets)
        spacing = [0] * stripes
        for idx in range(0, stripes - 1, 2):
            spacing[idx] = self.stripe_spacing
        for idx in range(1, stripes, 2):
            spacing[idx] = self.module_spacing
        # We don't want the final stripe to have a gap afterwards
        spacing[-1] = 0

        return spacing

    def create_vds_maps(self, source_meta):
        """Create a list of VirtualMaps of raw data to the VDS.

        Args:
            source_meta(SourceMeta): Source attributes

        Returns:
            list(VirtualMap): Maps describing links between raw data and VDS

        """
        source_shape = source_meta.frames + (source_meta.height,
                                             source_meta.width)
        spacing = self.construct_vds_spacing()
        target_height = source_meta.height * len(self.datasets) + sum(spacing)
        target_shape = source_meta.frames + (target_height,
                                             source_meta.width)
        self.logger.debug("VDS metadata:\n"
                          "  Shape: %s\n"
                          "  Spacing: %s", target_shape, spacing)

        vds = VirtualTarget(self.output_file, self.target_node,
                            shape=target_shape)

        map_list = []
        current_position = 0
        for idx, dataset in enumerate(self.datasets):
            v_source = VirtualSource(dataset, self.source_node,
                                     shape=source_shape)

            start = current_position
            stop = start + source_meta.height + spacing[idx]
            current_position = stop

            # Hyperslab: Full slice for frames for each axis,
            #            Height bounds of stripe[idx],
            #            Full slice for width
            hyperslab = tuple([self.FULL_SLICE] * len(source_meta.frames) +
                              [slice(start, stop)] +
                              [self.FULL_SLICE])

            v_target = vds[hyperslab]
            v_map = VirtualMap(v_source, v_target, dtype=source_meta.dtype)

            self.logger.debug("Mapping dataset %s to %s of %s.",
                              dataset.split("/")[-1], hyperslab, self.name)
            map_list.append(v_map)

        return map_list
